#include "joystick.h"
#include "pyrainput.h"
#include <funkeymonkey/evdevdevice.h>
#include <linux/input.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/select.h>
#define nBitsPerUchar			(sizeof(unsigned char) * 8)
#define nUcharsForNBits(nBits)	((((nBits)-1)/nBitsPerUchar)+1)
#define bitOffsetInUchar(bit)	((bit)%nBitsPerUchar)
#define ucharIndexForBit(bit)	((bit)/nBitsPerUchar)
#define testBit(bit, array)		((array[ucharIndexForBit(bit)] >> bitOffsetInUchar(bit)) & 1)

PyraJoystick::PyraJoystick(std::string const& path, unsigned int bus, std::string const& name, unsigned int vendor, unsigned int product, unsigned int version, std::vector<PossibleEvent> const& possibleEvents, std::vector<AbsoluteAxisCalibrationData> const& absoluteAxesCalibrationData) : UinputDevice(path, bus, name, vendor, product, version, possibleEvents, absoluteAxesCalibrationData, true) {
	std::string dev_path = "";
	std::vector<EvdevDevice::Information> vec = EvdevDevice::availableDevices();
	for(const auto& info: vec) {
		if (info.name == "twl6040:vibrator")
			dev_path = info.path;
	}
	if (dev_path!="")
		_rumble = open(dev_path.data(), O_WRONLY | O_NDELAY);
	ready = (dev_path!="" && _rumble>0);
	int cnt_effects = 0;
	if (ready) {
		ioctl(_rumble, EVIOCGEFFECTS, &cnt_effects);
		// Forwarding the supported effect declaration
		if (ioctl(_fd, UI_SET_EVBIT, EV_FF)>=0) {
			unsigned char features[1 + FF_MAX/8/sizeof(unsigned char)];
			int ret = ioctl(_rumble, EVIOCGBIT(EV_FF, sizeof(features)*sizeof(unsigned char)), features);
			auto effectList = {FF_RUMBLE, FF_PERIODIC, FF_RAMP, FF_CONSTANT, FF_SPRING, FF_FRICTION, FF_DAMPER, FF_INERTIA, FF_GAIN, FF_AUTOCENTER, FF_SQUARE, FF_TRIANGLE, FF_SINE, FF_SAW_UP, FF_SAW_DOWN};
			ioctl(_fd, UI_SET_EVBIT, EV_FF);
			for (const auto bit: effectList) {
				if (testBit(bit, features))
					ioctl(_fd, UI_SET_FFBIT, bit);
			}
		}
	}
	init(cnt_effects, bus, name, vendor, product, version, possibleEvents, absoluteAxesCalibrationData);
}

PyraJoystick::~PyraJoystick() {
	if (ready) {
		close(_rumble);
	}
}

bool PyraJoystick::is_ready() { 
	return ready;
}

void PyraJoystick::deleteEffect(int id) {
	if (_effectMap.find(id) != _effectMap.end()) {
		ioctl(_rumble, EVIOCRMFF, _effectMap[id]);
		_effectMap.erase(id);
	}
}

void PyraJoystick::rumble_handle() {
	struct input_event ev;
	int rv;
	if (! ready) return;
	int ret;
	while ((ret = read(_fd, &ev, sizeof(ev))) == sizeof(ev)) {
		switch(ev.type) {
		case EV_FF:
			switch(ev.code)	{
			case FF_GAIN:
				struct input_event gain;
				memset(&gain, 0, sizeof(gain));
				gain.type = EV_FF;
				gain.code = FF_GAIN;
				gain.value = ev.value;
				write(_rumble, &gain, sizeof(gain));
			break;
			default:
				if (ev.value) {
					int code  = ev.code;
					if (_effectMap.find(code) != _effectMap.end())
						code = _effectMap[code];
					struct input_event play;
					memset(&play, 0, sizeof(play));
					play.type = EV_FF;
					play.code = code;
					play.value = 1;
					int ret = write(_rumble, &play, sizeof(play));
				} else {
					int code  = ev.code;
					if (_effectMap.find(code) != _effectMap.end())
						code = _effectMap[code];
					struct input_event stop;
					memset(&stop, 0, sizeof(stop));
					stop.type = EV_FF;
					stop.code = code;
					stop.value = 0;
					int ret = write(_rumble, &stop, sizeof(stop));
				}
			break;
			}
		break;
		case EV_UINPUT:
			switch (ev.code) {
				case UI_FF_UPLOAD: {
					struct uinput_ff_upload upload;
					memset(&upload, 0, sizeof(upload));
					// *VERY* important, without this you break the kernel and have to reboot due to dead hanging process
					upload.request_id = ev.value;
					ioctl(_fd, UI_BEGIN_FF_UPLOAD, &upload);
					int orig = upload.effect.id;
					if (_effectMap.find(orig) != _effectMap.end()) {
						// effect already known, removing 1st
						deleteEffect(orig);
					}
					upload.effect.id = -1;
					ioctl(_rumble, EVIOCSFF, &upload.effect);
					_effectMap[orig] = upload.effect.id;
					upload.retval = 0;
					ioctl(_fd, UI_END_FF_UPLOAD, &upload);
				}
				break;
				case UI_FF_ERASE: {
					struct uinput_ff_erase erase;
					memset(&erase, 0, sizeof(erase));
					// *VERY* important, without this you break the kernel and have to reboot due to dead hanging process
					erase.request_id = ev.value;
					ioctl(_fd, UI_BEGIN_FF_ERASE, &erase);
					deleteEffect(erase.effect_id);
					erase.retval = 0;
					ioctl(_fd, UI_END_FF_ERASE, &erase);
				}
				break;
			}
			break;
		}
	}
}
