#ifndef PYRAINPUT_H
#define PYRAINPUT_H
#include <funkeymonkey/funkeymonkeymodule.h>
#include <condition_variable>
#include <functional>
#include <thread>
#include <unordered_map>
#include <mutex>

#include "joystick.h"
#include "KeyBehaviors.h"


#ifndef REL_WHEEL_HI_RES
#define REL_WHEEL_HI_RES 0x0b
#endif
#ifndef REL_HWHEEL_HI_RES
#define REL_HWHEEL_HI_RES 0x0c
#endif

enum Role { ROLE_LEFT_NUB, ROLE_RIGHT_NUB, ROLE_KEYBOARD, ROLE_GPIO };

struct Scripts {
	std::string brightnessNormal;
	std::string brightnessFn;
	std::string brightnessShift;
	std::string brightnessFnShift;
	std::string brightnessAlt;
	std::string brightnessCtrl;
	std::string brightnessAltCtrl;
	std::string brightnessFnAlt;
	std::string brightnessFnCtrl;
	std::string brightnessShiftAlt;
	std::string brightnessShiftCtrl;
	std::string brightnessFnShiftAlt;
	std::string brightnessFnShiftCtrl;
	std::string commandModeOn;
	std::string commandModeOff;
};

struct Mouse {
	// Any thread using device must hold mutex
	UinputDevice device;
	
	// May be changed from any thread at any time
	int dx;
	int dy;
	int dwx;
	int dwy;

	// Mouse wheel position
	int wheelPosition;
	int hwheelPosition;
	static int const WHEEL_HI_RES_PER_NOTCH = 120;

	// Used to signal changes in above values, uses mutex
	std::condition_variable signal;
	
	// Mouse device mutex
	std::mutex mutex;
};


struct Settings {
	enum NubAxisMode {
		UNKNOWN_NUB_AXIS_MODE,
		MOUSE_X, MOUSE_Y, MOUSE_BTN,
		SCROLL_X, SCROLL_Y
	};
	enum NubClickMode {
		UNKNOWN_NUB_CLICK_MODE,
		NUB_CLICK_LEFT, NUB_CLICK_RIGHT,
		MOUSE_LEFT, MOUSE_MIDDLE, MOUSE_RIGHT
	};
	enum AltMode {
		LEFT_ALT, RIGHT_ALT
	};
        enum ProfileIndex {
                PROFILE_0, PROFILE_1, PROFILE_2, PROFILE_3, PROFILE_4,
                PROFILE_5, PROFILE_6, PROFILE_7, PROFILE_8, PROFILE_9,
                PROFILE_NONE
        };
	
        std::vector<NubAxisMode> leftNubModeX = { MOUSE_X };
        std::vector<NubAxisMode> leftNubModeY = { MOUSE_Y };
        std::vector<NubAxisMode> rightNubModeX = { MOUSE_BTN };
        std::vector<NubAxisMode> rightNubModeY = { SCROLL_Y };
        std::vector<NubClickMode> leftNubClickMode = { MOUSE_LEFT };
        std::vector<NubClickMode> rightNubClickMode = { MOUSE_RIGHT };
    
        std::vector<AltMode> startButton  = { LEFT_ALT };
        std::vector<AltMode> rightTrigger = { RIGHT_ALT };
	
	// May be changed from any thread at any time
	int joyDeadzone = 10;
	int mouseDeadzone = 50;
	int mouseSensitivity = 75;
	int mouseWheelSensitivity = 75;
	int mouseWheelDeadzone = 100;
	int mouseClickDeadzone = 100;
	
	bool exportGamepad = true;
	bool exportMouse   = true;
	bool exportKeypad  = true;

	std::string configFile;
        std::string profiles[10];
        ProfileIndex currentProfile = PROFILE_NONE;
	Scripts scripts;
};

using AltModeMap = std::unordered_map<std::string, Settings::AltMode>;
AltModeMap const ALT_MODES = {
	{ "left_alt",  Settings::LEFT_ALT },
	{ "right_alt", Settings::RIGHT_ALT }
};

using NubAxisModeMap = std::unordered_map<std::string, Settings::NubAxisMode>;
NubAxisModeMap const NUB_AXIS_MODES = {
	{ "mouse_x", Settings::MOUSE_X },
	{ "mouse_y", Settings::MOUSE_Y },
	{ "mouse_btn", Settings::MOUSE_BTN },
	{ "scroll_x", Settings::SCROLL_X },
	{ "scroll_y", Settings::SCROLL_Y }
};

using NubClickModeMap = std::unordered_map<std::string, Settings::NubClickMode>;
NubClickModeMap const NUB_CLICK_MODES = {
	{ "nub_click_left", Settings::NUB_CLICK_LEFT },
	{ "nub_click_right", Settings::NUB_CLICK_RIGHT },
	{ "mouse_left", Settings::MOUSE_LEFT },
	{ "mouse_middle", Settings::MOUSE_MIDDLE },
	{ "mouse_right", Settings::MOUSE_RIGHT }
};

void handleArgs(char const** argv, unsigned int argc, Settings& settings);
void loadConfig(std::string const& filename, Settings& settings);
void loadProfile(Settings::ProfileIndex index, Settings& settings);
void setCommandMode(bool value);
Settings::NubAxisMode parseNubAxisMode(std::string const& str);
Settings::NubClickMode parseNubClickMode(std::string const& str);
Settings::AltMode parseAltMode(std::string const& str);
void handleNubAxis(Settings::NubAxisMode mode, int value, Mouse* mouse, PyraJoystick* gamepad, Settings const& settings);
void handleNubClick(Settings::NubClickMode mode, int value, Mouse* mouse, PyraJoystick* gamepad, Settings const& settings);
void handleBrightness(int value);
void resetSettings(Settings& settings);
// Mouse movement/scroll thread handler
void handleMouse(Mouse* mouse, Settings* settings, bool* stop);
void handleRumble(PyraJoystick* gamepad, bool* stop);

// Parse whitespace separated list of items
template<class T> std::vector<T> parseMultivalue(std::string const& str, std::function<T(std::string const&)> fn);

struct pyrainput_global {
	bool stop = false;
	PyraJoystick* gamepad = nullptr;
	UinputDevice* keyboard = nullptr;
	KeyBehaviors* behaviors = nullptr;
	Mouse* mouse = nullptr;
	std::thread mouseThread;
	std::thread rumbleThread;
	Settings settings;
	LeftRight Fn;
	LeftRight Alt;
	LeftRight Shift;
	LeftRight Ctrl;
	int hatx = 0;
	int haty = 0;
	int mouseBtn = 0;
	bool commandMode = false;
};

extern struct pyrainput_global global;

#endif
