#include "pyrainput.h"
#include <regex>
#include <iostream>
#include <fstream>

void resetSettings(Settings& settings) {
	loadProfile(Settings::PROFILE_NONE, global.settings);
	settings.leftNubModeX      = { Settings::MOUSE_X };
	settings.leftNubModeY      = { Settings::MOUSE_Y };
	settings.rightNubModeX     = { Settings::MOUSE_BTN };
	settings.rightNubModeY     = { Settings::SCROLL_Y };
	settings.leftNubClickMode  = { Settings::MOUSE_LEFT };
	settings.rightNubClickMode = { Settings::MOUSE_RIGHT };
	settings.startButton       = { Settings::LEFT_ALT };
	settings.rightTrigger      = { Settings::RIGHT_ALT};
	settings.joyDeadzone       = 10;
	settings.mouseDeadzone     = 50;
	settings.mouseSensitivity  = 75;
	settings.mouseWheelSensitivity = 75;
	settings.mouseWheelDeadzone = 100;
	settings.mouseClickDeadzone = 100;

	settings.exportGamepad = true;
	settings.exportMouse   = true;
	settings.exportKeypad  = true;
}

template<class T> std::vector<T> parseMultivalue(std::string const& str, std::function<T(std::string const&)> fn) {
	std::istringstream stream(str);
	std::vector<T> values;
	std::transform(std::istream_iterator<std::string>(stream), std::istream_iterator<std::string>(), std::back_inserter(values), fn);
	return values;
}

using SettingHandler = std::function<void(std::string const&,Settings&)>;
using SettingHandlerMap = std::unordered_map<std::string, SettingHandler>;
SettingHandlerMap const SETTING_HANDLERS = {
	{ "scripts.brightness.normal", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessNormal = value;
	} },
	{ "scripts.brightness.fn", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFn = value;
	} },
	{ "scripts.brightness.shift", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessShift = value;
	} },
	{ "scripts.brightness.fnshift", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnShift = value;
	} },
	{ "scripts.brightness.alt", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessAlt = value;
	} },
	{ "scripts.brightness.ctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessCtrl = value;
	} },
	{ "scripts.brightness.altctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessAltCtrl = value;
	} },
	{ "scripts.brightness.fnalt", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnAlt = value;
	} },
	{ "scripts.brightness.fnctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnCtrl = value;
	} },
	{ "scripts.brightness.shiftalt", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessShiftAlt = value;
	} },
	{ "scripts.brightness.shiftctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessShiftCtrl = value;
	} },
	{ "scripts.brightness.fnshiftalt", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnShiftAlt = value;
	} },
	{ "scripts.brightness.fnshiftctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnShiftCtrl = value;
	} },
	{ "scripts.commandmode.on", [](std::string const& value, Settings& settings){
		settings.scripts.commandModeOn = value;
	} },
	{ "scripts.commandmode.off", [](std::string const& value, Settings& settings){
		settings.scripts.commandModeOff = value;
	} },
	{ "profile.0", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_0] = value;
	} },
	{ "profile.1", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_1] = value;
	} },
	{ "profile.2", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_2] = value;
	} },
	{ "profile.3", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_3] = value;
	} },
	{ "profile.4", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_4] = value;
	} },
	{ "profile.5", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_5] = value;
	} },
	{ "profile.6", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_6] = value;
	} },
	{ "profile.7", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_7] = value;
	} },
	{ "profile.8", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_8] = value;
	} },
	{ "profile.9", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_9] = value;
	} },
	{ "profile.current", [](std::string const& value, Settings& settings){
		int intValue = std::stoi(value);

		if(intValue >= 0 && intValue < 10) {
			loadProfile(static_cast<Settings::ProfileIndex>(intValue), settings);
		}
	} },
	{ "gamepad.export", [](std::string const& value, Settings& settings){
		settings.exportGamepad = (value != "0");
	} },
	{ "keypad.export", [](std::string const& value, Settings& settings){
		settings.exportKeypad = (value != "0");
	} },
	{ "mouse.export", [](std::string const& value, Settings& settings){
		settings.exportMouse = (value != "0");
	} },
	{ "mouse.sensitivity", [](std::string const& value, Settings& settings){
		settings.mouseSensitivity = std::stoi(value);
	} },
	{ "mouse.deadzone", [](std::string const& value, Settings& settings) {
		settings.mouseDeadzone = std::stoi(value);
	} },
	{ "mouse.wheel.sensitivity", [](std::string const& value, Settings& settings){
		settings.mouseWheelSensitivity = std::stoi(value);
	} },
	{ "mouse.wheel.deadzone", [](std::string const& value, Settings& settings) {
		settings.mouseWheelDeadzone = std::stoi(value);
	} },
	{ "mouse.click.deadzone", [](std::string const& value, Settings& settings) {
		settings.mouseClickDeadzone = std::stoi(value);
	} },
	{ "nubs.deadzone", [](std::string const& value, Settings& settings) {
		settings.joyDeadzone = std::stoi(value);
	} },
	{ "nubs.left.x", [](std::string const& value, Settings& settings) {
                settings.leftNubModeX = parseMultivalue<Settings::NubAxisMode>(value, parseNubAxisMode);
	} },
	{ "nubs.left.y", [](std::string const& value, Settings& settings) {
		settings.leftNubModeY = parseMultivalue<Settings::NubAxisMode>(value, parseNubAxisMode);
	} },
	{ "nubs.right.x", [](std::string const& value, Settings& settings) {
		settings.rightNubModeX = parseMultivalue<Settings::NubAxisMode>(value, parseNubAxisMode);
	} },
	{ "nubs.right.y", [](std::string const& value, Settings& settings) {
		settings.rightNubModeY = parseMultivalue<Settings::NubAxisMode>(value, parseNubAxisMode);
	} },
	{ "nubs.left.click", [](std::string const& value, Settings& settings) {
		settings.leftNubClickMode = parseMultivalue<Settings::NubClickMode>(value, parseNubClickMode);
	} },
	{ "nubs.right.click", [](std::string const& value, Settings& settings) {
		settings.rightNubClickMode = parseMultivalue<Settings::NubClickMode>(value, parseNubClickMode);
	} },
	{ "altmode.start.button", [](std::string const& value, Settings& settings) {
		settings.startButton = parseMultivalue<Settings::AltMode>(value, parseAltMode);
	} },
	{ "altmode.right.trigger", [](std::string const& value, Settings& settings) {
		settings.rightTrigger = parseMultivalue<Settings::AltMode>(value, parseAltMode);
	} }
};

void loadConfig(std::string const& filename, Settings& settings) {
	std::regex re("^([\\w.]+)\\s*=\\s*(.*)$");
	std::regex emptyRe("^\\s*$");
	std::ifstream configFile(filename);
	
	if(!configFile) {
		std::cerr << "ERROR: Could not open config file " << filename << std::endl;
		return;
	}
	std::string line;
	while(std::getline(configFile, line)) {
		if(line.empty() || line.at(0) == '#' || std::regex_match(line, emptyRe))
			continue;

		std::smatch match;
		if(std::regex_match(line, match, re)) {
			std::string key = match[1];
			std::string value = match[2];
			std::transform(key.begin(), key.end(), key.begin(), tolower);
			auto iter = SETTING_HANDLERS.find(key);
			if(iter == SETTING_HANDLERS.end()) {
				std::cout << "WARNING: Unknown setting in config file: " 
				<< key << std::endl;
			} else {
				iter->second(value, settings);
			}
		} else {
			std::cerr << "Invalid line in config file: " << line << std::endl;
		}
	}
	/* Apply "Alt Keys" mapping configuration */
    if (std::find(settings.startButton.begin(), settings.startButton.end(), Settings::LEFT_ALT) != settings.startButton.end()) {
        global.behaviors->gpmap2(KEY_LEFTALT,	BTN_START, &global.Alt.left, &global.Alt);
    }
    if (std::find(settings.startButton.begin(), settings.startButton.end(), Settings::RIGHT_ALT) != settings.startButton.end()) {
        global.behaviors->gpmap2(KEY_RIGHTALT,	BTN_START, &global.Alt.right, &global.Alt);
    }
    if (std::find(settings.rightTrigger.begin(), settings.rightTrigger.end(), Settings::LEFT_ALT) != settings.rightTrigger.end()) {
        global.behaviors->gpmap2(KEY_LEFTALT,	BTN_TR2, &global.Alt.left, &global.Alt);
    }
    if (std::find(settings.rightTrigger.begin(), settings.rightTrigger.end(), Settings::RIGHT_ALT) != settings.rightTrigger.end()) {
        global.behaviors->gpmap2(KEY_RIGHTALT,	BTN_TR2, &global.Alt.right, &global.Alt);
    }
}

void loadProfile(Settings::ProfileIndex index, Settings& settings) {
  if(index != settings.currentProfile) {
    if(index != Settings::PROFILE_NONE) {
      std::string const& profilePath = settings.profiles[index];
      loadConfig(profilePath, settings);
    }
    settings.currentProfile = index;
  }
}

Settings::NubAxisMode parseNubAxisMode(std::string const& str) {
	std::string s(str);
	std::transform(s.begin(), s.end(), s.begin(), tolower);
	auto iter = NUB_AXIS_MODES.find(s);
	if(iter == NUB_AXIS_MODES.end()) {
		return Settings::UNKNOWN_NUB_AXIS_MODE;
	} else {
		return iter->second;
	}
}

Settings::NubClickMode parseNubClickMode(std::string const& str) {
	std::string s(str);
	std::transform(s.begin(), s.end(), s.begin(), tolower);
	auto iter = NUB_CLICK_MODES.find(s);
	if(iter == NUB_CLICK_MODES.end()) {
		return Settings::UNKNOWN_NUB_CLICK_MODE;
	} else {
		return iter->second;
	}
}

Settings::AltMode parseAltMode(std::string const& str) {
	std::string s(str);
	std::transform(s.begin(), s.end(), s.begin(), tolower);
	auto iter = ALT_MODES.find(s);
	if(iter == ALT_MODES.end()) {
		return Settings::LEFT_ALT;
	} else {
		return iter->second;
	}
}
